@@ -3,7 +3,7 @@ package autodie::hints;
 use strict;
 use warnings;
 
-our $VERSION = '1.999_01';
+our $VERSION = '2.00';
 
 =head1 NAME
 
@@ -11,30 +11,54 @@ autodie::hints - Provide hints about user subroutines to autodie
 
 =head1 SYNOPSIS
 
-   package Your::Module;
+    package Your::Module;
 
-   sub AUTODIE_HINTS {
-       return {
-           foo => { scalar => HINTS, list => SOME_HINTS },
-           bar => { scalar => HINTS, list => MORE_HINTS },
-       }
-   }
+    our %DOES = ( 'autodie::hints::provider' => 1 );
+
+    sub AUTODIE_HINTS {
+        return {
+            foo => { scalar => HINTS, list => SOME_HINTS },
+            bar => { scalar => HINTS, list => MORE_HINTS },
+        }
+    }
+
+    # Later, in your main program...
+
+    use Your::Module qw(foo bar);
+    use autodie      qw(:default foo bar);
+
+    foo();         # succeeds or dies based on scalar hints
+
+    # Alternatively, hints can be set on subroutines we've
+    # imported.
+
+    use autodie::hints;
+    use Some::Module qw(think_positive);
+
+    BEGIN {
+        autodie::hints->set_hints_for(
+            \&think_positive,
+            {
+                fail => sub { $_[0] <= 0 }
+            }
+        )
+    }
+    use autodie qw(think_positive);
+
+    think_positive(...);    # Returns positive or dies.
 
 
-   # later
-   use Your::Module qw(foo bar);
-   use autodie      qw(:default foo bar);
+=head1 DESCRIPTION
 
-   foo();         # succeeds or dies based on scalar hints
-   print foo();   # succeeds or dies based on list hints
+=head2 Introduction
 
-=head1 Hinting interface
+The L<autodie> pragma is very smart when it comes to working with
+Perl's built-in functions.  The behaviour for these functions are
+fixed, and C<autodie> knows exactly how they try to signal failure.
 
-C<autodie::hints> allows you to tell C<autodie> what your subroutines
-return on failure.
-
-Without hints, C<autodie> only considers the following return values as
-evidence of failure:
+But what about user-defined subroutines from modules?  If you use
+C<autodie> on a user-defined subroutine then it assumes the following
+behaviour to demonstrate failure:
 
 =over
 
@@ -53,17 +77,46 @@ A list containing a single undef, in list context
 =back
 
 All other return values (including the list of the single zero, and the
-list containing a single empty string) are considered true.  However,
+list containing a single empty string) are considered successful.  However,
 real-world code isn't always that easy.  Perhaps the code you're working
 with returns a string containing the word "FAIL" in it upon failure, or a
 two element list containing C<(undef, "human error message")>.  To make
-autodie work with these, we have the hinting interface.
+autodie work with these sorts of subroutines, we have
+the I<hinting interface>.
+
+The hinting interface allows I<hints> to be provided to C<autodie>
+on how it should detect failure from user-defined subroutines.  While
+these I<can> be provided by the end-user of C<autodie>, they are ideally
+written into the module itself, or into a helper module or sub-class
+of C<autodie> itself.
+
+=head2 What are hints?
+
+A I<hint> is a subroutine or value that is checked against the
+return value of an autodying subroutine.  If the match returns true,
+C<autodie> considers the subroutine have failed.
+
+If the hint provided is a subroutine, then C<autodie> will pass
+the complete return value to that subroutine.  If the hint is
+any other value, then C<autodie> will smart-match against the
+value provided.  In Perl 5.8.x, there is no smart-match operator, and as such
+only subroutine hints are supported in these versions.
+
+Hints can be provided for both scalar context and list context.  Note
+that an autodying subroutine will never see a void context, as
+C<autodie> always needs to capture the return value for examination.
+Autodying subroutines called in void context act as if they're called
+in a scalar context, but their return value is discarded after it
+has been checked.
 
 =head2 Example hints
 
-Hints may consist of scalars, array references, regular expression and
-subroutine references.  You can specify different hints for how failure should
-be identified in scalar and list contexts.
+Hints may consist of scalars, array references, regular expressions and
+subroutine references.  You can specify different hints for how
+failure should be identified in scalar and list contexts.
+
+These examples apply for use in the C<AUTODIE_HINTS> subroutine and when
+calling C<autodie::hints->set_hints_for()>.
 
 The most common context-specific hints are:
 
@@ -79,10 +132,10 @@ The most common context-specific hints are:
         # List failures always return empty list:
             {  list => []  }
 
-        # List failures return C<()> or C<(undef)> [default expectation]:
+        # List failures return () or (undef) [default expectation]:
             {  list => sub { ! @_ || @_ == 1 && !defined $_[0] }  }
 
-        # List failures return C<()> or a single false value:
+        # List failures return () or a single false value:
             {  list => sub { ! @_ || @_ == 1 && !$_[0] }  }
 
         # List failures return (undef, "some string")
@@ -108,14 +161,14 @@ The most common context-specific hints are:
 	);
 
 This "in all contexts" construction is very common, and can be
-abbreviated, using the 'fail' key. A C<< { fail => $val } >> hint is
-simply a shortcut for C<< { scalar => $val, list => [ $val ] } >>:
+abbreviated, using the 'fail' key.  This sets both the C<scalar>
+and C<list> hints to the same value:
 
         # Unsuccessful foo() returns 0 in all contexts...
         autodie::hints->set_hints_for(
             \&foo,
             {
-                fail => 0
+                fail => sub { @_ == 1 and defined $_[0] and $_[0] == 0 }
             }
 	);
 
@@ -135,20 +188,15 @@ simply a shortcut for C<< { scalar => $val, list => [ $val ] } >>:
             }
 	);
 
-        # Unsuccessful bizarro_system() returns random value and sets $?...
-        autodie::hints->set_hints_for(
-            \&bizarro_system,
-            {
-                fail => sub { defined $? }
-            }
-	);
+=head1 Manually setting hints from within your program
 
-On Perl 5.8, only simple scalars, array references, regular expressions and
-subroutines are supported as hints, anything else is a compile-time error.
+If you are using a module which returns something special on failure, then
+you can manually create hints for each of the desired subroutines.  Once
+the hints are specified, they are available for all files and modules loaded
+thereafter, thus you can move this work into a module and it will still
+work.
 
-=head1 Setting hints directly
-
-	package Your::Module;
+	use Some::Module qw(foo bar);
 	use autodie::hints;
 
 	autodie::hints->set_hints_for(
@@ -158,53 +206,116 @@ subroutines are supported as hints, anything else is a compile-time error.
 			list   => LIST_HINT,
 		}
 	);
+	autodie::hints->set_hints_for(
+		\&bar,
+                { fail => SOME_HINT, }
+	);
 
 It is possible to pass either a subroutine reference (recommended) or a fully
-qualified subroutine name as the first argument, so you can set hints on
-modules that I<might> get loaded, but haven't been loaded yet.
+qualified subroutine name as the first argument.  This means you can set hints
+on modules that I<might> get loaded:
 
-The hints above are smart-matched against the return value from the
-subroutine; a true result indicates failure, and an appropriate exception is
-thrown.  Since one can smart-match against a subroutine, it's possible to do
-quite complex checks for failure if needed.
+	use autodie::hints;
+	autodie::hints->set_hints_for(
+		'Some::Module:bar', { fail => SCALAR_HINT, }
+	);
 
-The hint-setting interface is pretty verbose, and is designed as something
-which might be written into sub-classes (my::company::autodie), or modules
-(preferably next to the subroutines themselves). 
+This technique is most useful when you have a project that uses a
+lot of third-party modules.  You can define all your possible hints
+in one-place.  This can even be in a sub-class of autodie.  For
+example:
 
-=head1 Auto-finding hints
+        package my::autodie;
+
+        use parent qw(autodie);
+        use autodie::hints;
+
+        autodie::hints->set_hints_for(...);
+
+        1;
+
+You can now C<use my::autodie>, which will work just like the standard
+C<autodie>, but is now aware of any hints that you've set.
+
+=head1 Adding hints to your module
+
+C<autodie> provides a passive interface to allow you to declare hints for
+your module.  These hints will be found and used by C<autodie> if it
+is loaded, but otherwise have no effect (or dependencies) without autodie.
+To set these, your module needs to declare that it I<does> the
+C<autodie::hints::provider> role.  This can be done by writing your
+own C<DOES> method, using a system such as C<Class::DOES> to handle
+the heavy-lifting for you, or declaring a C<%DOES> package variable
+with a C<autodie::hints::provider> key and a corresponding true value.
+
+Note that checking for a C<%DOES> hash is an C<autodie>-only
+short-cut.  Other modules do not use this mechanism for checking
+roles, although you can use the C<Class::DOES> module from the
+CPAN to allow it.
+
+In addition, you must define a C<AUTODIE_HINTS> subroutine that returns
+a hash-reference containing the hints for your subroutines:
 
 	package Your::Module;
 
+        # We can use the Class::DOES from the CPAN to declare adherence
+        # to a role.
+
+        use Class::DOES 'autodie::hints::provider' => 1;
+
+        # Alternatively, we can declare the role in %DOES.  Note that
+        # this is an autodie specific optimisation, although Class::DOES
+        # can be used to promote this to a true role declaration.
+
+        our %DOES = ( 'autodie::hints::provider' => 1 );
+
+        # Finally, we must define the hints themselves.
+
 	sub AUTODIE_HINTS {
 	    return {
 	        foo => { scalar => HINTS, list => SOME_HINTS },
 	        bar => { scalar => HINTS, list => MORE_HINTS },
+	        baz => { fail => HINTS },
 	    }
 	}
 
 This allows your code to set hints without relying on C<autodie> and
-C<autodie::hints>.  Thus if your end user chooses to use C<autodie> then hints
-declared in this way will be found and loaded for correct error handling.
+C<autodie::hints> being loaded, or even installed.  In this way your
+code can do the right thing when C<autodie> is installed, but does not
+need to depend upon it to function.
 
 =head1 Insisting on hints
 
+When a user-defined subroutine is wrapped by C<autodie>, it will
+use hints if they are available, and otherwise reverts to the
+I<default behaviour> described in the introduction of this document.
+This can be problematic if we expect a hint to exist, but (for
+whatever reason) it has not been loaded.
+
+We can ask autodie to I<insist> that a hint be used by prefixing
+an exclamation mark to the start of the subroutine name.  A lone
+exclamation mark indicates that I<all> subroutines after it must
+have hints declared.
+
 	# foo() and bar() must have their hints defined
 	use autodie qw( !foo !bar baz );
 
-	# Everything must have hints.
+	# Everything must have hints (recommended).
 	use autodie qw( ! foo bar baz );
 
 	# bar() and baz() must have their hints defined
 	use autodie qw( foo ! bar baz );
 
-It is possible for a user to insist that hints have been defined.  This is
-done by prefixing each user-defined subroutine with a C<!> in the import
-list.  A C<!> on its own specifies that all user-defined subroutines after
-that point must have hints.
+        # Enable autodie for all of Perl's supported built-ins,
+        # as well as for foo(), bar() and baz().  Everything must
+        # have hints.
+        use autodie qw( ! :all foo bar baz );
 
 If hints are not available for the specified subroutines, this will cause a
-compile-time error.
+compile-time error.  Insisting on hints for Perl's built-in functions
+(eg, C<open> and C<close>) is always successful.
+
+Insisting on hints is I<strongly> recommended.
 
 =cut
 
@@ -288,7 +399,12 @@ sub load_hints {
 
     my ($package) = ( $sub =~ /(.*)::/ );
 
-    # TODO: What do we do if we can't find a package?
+    if (not defined $package) {
+        require Carp;
+        Carp::croak(
+            "Internal error in autodie::hints::load_hints - no package found.
+        ");
+    }
 
     # Do nothing if we've already tried to load hints for
     # this package.
@@ -296,11 +412,15 @@ sub load_hints {
 
     my $hints_available = 0;
 
-    if ($package->can('DOES') and $package->DOES(HINTS_PROVIDER) ) {
-        $hints_available = 1;
-    }
-    elsif ( $package->isa(HINTS_PROVIDER) ) {
-        $hints_available = 1;
+    {
+        no strict 'refs';   ## no critic
+
+        if ($package->can('DOES') and $package->DOES(HINTS_PROVIDER) ) {
+            $hints_available = 1;
+        }
+        elsif ( ${"${package}::DOES"}{HINTS_PROVIDER.""} ) {
+            $hints_available = 1;
+        }
     }
 
     return if not $hints_available;
@@ -416,13 +536,13 @@ sub set_hints_for {
 __END__
 
 
-
-
 =head1 Diagnostics
 
-=head2 Attempts to set_hints_for unidentifiable subroutine
+=over 4
 
-You've called C<autodie::hints->set_hints_for()> using a subroutine
+=item Attempts to set_hints_for unidentifiable subroutine
+
+You've called C<< autodie::hints->set_hints_for() >> using a subroutine
 reference, but that reference could not be resolved back to a
 subroutine name.  It may be an anonymous subroutine (which can't
 be made autodying), or may lack a name for other reasons.
@@ -431,6 +551,20 @@ If you receive this error with a subroutine that has a real name,
 then you may have found a bug in autodie.  See L<autodie/BUGS>
 for how to report this.
 
+=item fail hints cannot be provided with either scalar or list hints for %s
+
+When defining hints, you can either supply both C<list> and
+C<scalar> keywords, I<or> you can provide a single C<fail> keyword.
+You can't mix and match them.
+
+=item %s hint missing for %s
+
+You've provided either a C<scalar> hint without supplying
+a C<list> hint, or vice-versa.  You I<must> supply both C<scalar>
+and C<list> hints, I<or> a single C<fail> hint.
+
+=back
+
 =head1 ACKNOWLEDGEMENTS
 
 =over 
@@ -458,6 +592,6 @@ same terms as Perl itself.
 
 =head1 SEE ALSO
 
-L<autodie>
+L<autodie>, L<Class::DOES>
 
 =cut
